/*   RCDCap
 *   Copyright (C) 2012  Zdravko Velinov
 *
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _RCDCAP_MEMORY_HH_
#define _RCDCAP_MEMORY_HH_

#include <memory>

#ifdef _WIN32
#   ifdef __MINGW32__
#       define _aligned_malloc __mingw_aligned_malloc
#       define _aligned_free __mingw_aligned_free
#   endif
#   include <malloc.h>
#   define RCDCAP_ALIGNED_ALLOC(ptr, nbytes, align) ptr = _aligned_malloc(nbytes, align)
#   define RCDCAP_ALIGNED_DEALLOC(ptr) _aligned_free(ptr)
#else
#   include <stdlib.h>
#   define  RCDCAP_ALIGNED_ALLOC(ptr, nbytes, align) if(posix_memalign(&ptr, align, nbytes)) ptr = 0;
#   define  RCDCAP_ALIGNED_DEALLOC(ptr) free(ptr)
#endif

namespace RCDCap
{
//! Returns next multiple of base.
inline size_t round(size_t n, size_t base)
{
    return ((n + base - 1)/base)*base;
}

/*! An aligned memory object allocator.
 *  \tparam T           the type of the object.
 *  \tparam alignment   the requested memory alignment.
 */
template<class T, int aligment>
struct ObjectAllocator
{
    //! Allocates enough memory to fit the object.
    static inline T* allocate()
    {
        void* ptr;
        RCDCAP_ALIGNED_ALLOC(ptr, sizeof(T), aligment);
        return reinterpret_cast<T*>(ptr);
    }

    //! Deallocates the already allocated memory.
    static inline void deallocate(T* ptr)
    {
        ptr->~T();
        RCDCAP_ALIGNED_DEALLOC(ptr);
    }
};

/*! \brief Deallocates an object aligned in the memory.
 *  \tparam T   the type of the object.
 */
template<class T>
struct AlignedDeleter
{
    //! Deallocates the memory.
    void operator()(T* p)
    {
        RCDCAP_ALIGNED_DEALLOC(p);
    }
};

#ifndef _MSC_VER
/*! Convenience function; it is included because there is no such function in the new standard,
 *  which is totally understandable, because it does not provide any benefits whatsoever.
 *  \tparam TClass      the name of the type for which an unique_ptr must be constructed.
 *  \tparam TArgs       the type of the arguments that are going to be passed to the constructor.
 *  \param args         the actual arguments that are going to be passed.
 *  \return an unique_ptr that is representing the object that was constructed.
 */
template<class TClass, class... TArgs>
std::unique_ptr<TClass> make_unique(TArgs&& ... args)
{
    return std::unique_ptr<TClass>(new TClass(args...));
}

/*! Convenience function; it is included because there is no such function in the new standard,
 *  which is totally understandable, because it does not provide any benefits whatsoever.
 *  \tparam TClass      the name of the type for which an unique_ptr must be constructed.
 *  \tparam TDel        the type of the deleter that is going to be used by the unique_ptr.
 *  \tparam TArgs       the type of the arguments that are going to be passed to the constructor.
 *  \param args         the actual arguments that are going to be passed.
 *  \return an unique_ptr that is representing the object that was constructed.
 */
template<class TClass, class TDel, class... TArgs>
std::unique_ptr<TClass, TDel> make_unique(TArgs&& ... args)
{
    return std::unique_ptr<TClass, TDel>(new TClass(args...));
}

/*! Convenience function; it is included because there is no such function in the new standard,
 *  which is totally understandable, because it does not provide any benefits whatsoever.
 *  \tparam TClass      the name of the type for which an unique_ptr must be constructed.
 *  \tparam TArgs       the type of the arguments that are going to be passed to the constructor.
 *  \param args         the actual arguments that are going to be passed.
 *  \return an unique_ptr that is representing the object that was constructed.
 */
template<class TClass, size_t alignment, class... TArgs>
std::unique_ptr<TClass, AlignedDeleter<TClass>> make_unique_aligned(TArgs&& ... args)
{
    return std::unique_ptr<TClass, AlignedDeleter<TClass>>(new (ObjectAllocator<TClass, alignment>::allocate()) TClass(args...), AlignedDeleter<TClass>());
}

#else
// <rant>
// And that is how it used to be; I'm keeping it because the MSVC team has decided that it
// would be cooler, if they implement tons of non-standard extensions(the MS way of beating
// the competition), instead of implementing the complete C++11 standard for the next version
// of VS.
// </rant>

template<class TClass>
std::unique_ptr<TClass> make_unique()
{
    return std::unique_ptr<TClass>(new TClass);
}

template<class TClass, class TArg1>
std::unique_ptr<TClass> make_unique(TArg1&& _arg1)
{
    return std::unique_ptr<TClass>(new TClass(_arg1));
}

template<class TClass, class TArg1, class TArg2>
std::unique_ptr<TClass> make_unique(TArg1&& _arg1, TArg2&& _arg2)
{
    return std::unique_ptr<TClass>(new TClass(_arg1));
}

template<class TClass, class TArg1, class TArg2, class TArg3>
std::unique_ptr<TClass> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3)
{
    return std::unique_ptr<TClass>(new TClass(_arg1, _arg2, _arg3));
}

template<class TClass, class TArg1, class TArg2, class TArg3, class TArg4>
std::unique_ptr<TClass> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3, TArg4&& _arg4)
{
    return std::unique_ptr<TClass>(new TClass(_arg1, _arg2, _arg3, _arg4));
}

template<class TClass, class TArg1, class TArg2, class TArg3, class TArg4, class TArg5>
std::unique_ptr<TClass> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3, TArg4&& _arg4, TArg5&& _arg5)
{
    return std::unique_ptr<TClass>(new TClass(_arg1, _arg2, _arg3, _arg4, _arg5));
}

template<class TClass, class TArg1, class TArg2, class TArg3, class TArg4, class TArg5,
                       class TArg6>
std::unique_ptr<TClass> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3, TArg4&& _arg4, TArg5&& _arg5,
                                    TArg6&& _arg6)
{
    return std::unique_ptr<TClass>(new TClass(_arg1, _arg2, _arg3, _arg4, _arg5,
                                              _arg6));
}

template<class TClass, class TArg1, class TArg2, class TArg3, class TArg4, class TArg5,
                       class TArg6, class TArg7>
std::unique_ptr<TClass> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3, TArg4&& _arg4, TArg5&& _arg5,
                                    TArg6&& _arg6, TArg7&& _arg7)
{
    return std::unique_ptr<TClass>(new TClass(_arg1, _arg2, _arg3, _arg4, _arg5,
                                              _arg6, _arg7));
}

template<class TClass, class TArg1, class TArg2, class TArg3, class TArg4, class TArg5,
                       class TArg6, class TArg7, class TArg8>
std::unique_ptr<TClass> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3, TArg4&& _arg4, TArg5&& _arg5,
                                    TArg6&& _arg6, TArg7&& _arg7, TArg8&& _arg8)
{
    return std::unique_ptr<TClass>(new TClass(_arg1, _arg2, _arg3, _arg4, _arg5,
                                              _arg6, _arg7, _arg8));
}

template<class TClass, class TArg1, class TArg2, class TArg3, class TArg4, class TArg5,
                       class TArg6, class TArg7, class TArg8, class TArg9>
std::unique_ptr<TClass> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3, TArg4&& _arg4, TArg5&& _arg5,
                                    TArg6&& _arg6, TArg7&& _arg7, TArg8&& _arg8, TArg9&& _arg9)
{
    return std::unique_ptr<TClass>(new TClass(_arg1, _arg2, _arg3, _arg4, _arg5,
                                              _arg6, _arg7, _arg8, _arg9));
}

template<class TClass, class TArg1, class TArg2, class TArg3, class TArg4, class TArg5,
                       class TArg6, class TArg7, class TArg8, class TArg9, class TArg10>
std::unique_ptr<TClass> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3, TArg4&& _arg4, TArg5&& _arg5,
                                    TArg6&& _arg6, TArg7&& _arg7, TArg8&& _arg8, TArg9&& _arg9, TArg10&& _arg10)
{
    return std::unique_ptr<TClass>(new TClass(_arg1, _arg2, _arg3, _arg4, _arg5,
                                              _arg6, _arg7, _arg8, _arg9, _arg10));
}

template<class TClass, class TDel>
std::unique_ptr<TClass, TDel> make_unique()
{
    return std::unique_ptr<TClass, TDel>(new TClass);
}

template<class TClass, class TDel, class TArg1>
std::unique_ptr<TClass, TDel> make_unique(TArg1&& _arg1)
{
    return std::unique_ptr<TClass, TDel>(new TClass(_arg1));
}

template<class TClass, class TDel, class TArg1, class TArg2>
std::unique_ptr<TClass, TDel> make_unique(TArg1&& _arg1, TArg2&& _arg2)
{
    return std::unique_ptr<TClass, TDel>(new TClass(_arg1));
}

template<class TClass, class TDel, class TArg1, class TArg2, class TArg3>
std::unique_ptr<TClass, TDel> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3)
{
    return std::unique_ptr<TClass, TDel>(new TClass(_arg1, _arg2, _arg3));
}

template<class TClass, class TDel, class TArg1, class TArg2, class TArg3, class TArg4>
std::unique_ptr<TClass, TDel> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3, TArg4&& _arg4)
{
    return std::unique_ptr<TClass, TDel>(new TClass(_arg1, _arg2, _arg3, _arg4));
}

template<class TClass, class TDel, class TArg1, class TArg2, class TArg3, class TArg4, class TArg5>
std::unique_ptr<TClass, TDel> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3, TArg4&& _arg4, TArg5&& _arg5)
{
    return std::unique_ptr<TClass, TDel>(new TClass(_arg1, _arg2, _arg3, _arg4, _arg5));
}

template<class TClass, class TDel, class TArg1, class TArg2, class TArg3, class TArg4, class TArg5,
                                   class TArg6>
std::unique_ptr<TClass, TDel> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3, TArg4&& _arg4, TArg5&& _arg5,
                                          TArg6&& _arg6)
{
    return std::unique_ptr<TClass, TDel>(new TClass(_arg1, _arg2, _arg3, _arg4, _arg5,
                                                    _arg6));
}

template<class TClass, class TDel, class TArg1, class TArg2, class TArg3, class TArg4, class TArg5,
                                   class TArg6, class TArg7>
std::unique_ptr<TClass, TDel> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3, TArg4&& _arg4, TArg5&& _arg5,
                                          TArg6&& _arg6, TArg7&& _arg7)
{
    return std::unique_ptr<TClass, TDel>(new TClass(_arg1, _arg2, _arg3, _arg4, _arg5,
                                                    _arg6, _arg7));
}

template<class TClass, class TDel, class TArg1, class TArg2, class TArg3, class TArg4, class TArg5,
                                   class TArg6, class TArg7, class TArg8>
std::unique_ptr<TClass, TDel> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3, TArg4&& _arg4, TArg5&& _arg5,
                                          TArg6&& _arg6, TArg7&& _arg7, TArg8&& _arg8)
{
    return std::unique_ptr<TClass, TDel>(new TClass(_arg1, _arg2, _arg3, _arg4, _arg5,
                                                    _arg6, _arg7, _arg8));
}

template<class TClass, class TDel, class TArg1, class TArg2, class TArg3, class TArg4, class TArg5,
                                   class TArg6, class TArg7, class TArg8, class TArg9>
std::unique_ptr<TClass, TDel> make_unique(TArg1 _arg1, TArg2 _arg2, TArg3 _arg3, TArg4 _arg4, TArg5 _arg5,
                                          TArg6 _arg6, TArg7 _arg7, TArg8 _arg8, TArg9 _arg9)
{
    return std::unique_ptr<TClass, TDel>(new TClass(_arg1, _arg2, _arg3, _arg4, _arg5,
                                                    _arg6, _arg7, _arg8, _arg9));
}

template<class TClass, class TDel, class TArg1, class TArg2, class TArg3, class TArg4, class TArg5,
                                   class TArg6, class TArg7, class TArg8, class TArg9, class TArg10>
std::unique_ptr<TClass, TDel> make_unique(TArg1&& _arg1, TArg2&& _arg2, TArg3&& _arg3, TArg4&& _arg4, TArg5&& _arg5,
                                          TArg6&& _arg6, TArg7&& _arg7, TArg8&& _arg8, TArg9&& _arg9, TArg10&& _arg10)
{
    return std::unique_ptr<TClass, TDel>(new TClass(_arg1, _arg2, _arg3, _arg4, _arg5,
                                                    _arg6, _arg7, _arg8, _arg9, _arg10));
}
#endif
}

#endif /* _RCDCAP_MEMORY_HH_ */